#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <errno.h>
#include "common.h"
#include "dbms.h"
#include "charset.h"
#include "proto.h"
#include "parseurl.h"
#include "parser.h"
#include "spell.h"
#include "conf.h"
#include "indexer.h"
#include "log.h"


char ExtraHeaders[STRSIZ]="";
char UserAgent[STRSIZ]="";
int MaxDocSize=MAXDOCSIZE;


static SERVER		*	Server=NULL;
static UDM_FILTER	*	Filter=NULL;
static ALIAS		*	Alias=NULL;

static int nservers=0,mservers=0;
static int nfilters=0,mfilters=0;
static int naliases=0,maliases=0;


/* One server structure to be available on all config levels*/
/* to pass parameters */
static SERVER * csrv=NULL;


/************ Allow/Disallow/CheckOnly  stuff *****************/


static int AddFilter(char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char regerrstr[STRSIZ]="";
int err;
	if(nfilters>=mfilters){
		if(mfilters){
			mfilters+=16;
			Filter=(UDM_FILTER *)realloc(Filter,mfilters*sizeof(UDM_FILTER));
		}else{
			mfilters=16;
			Filter=(UDM_FILTER *)malloc(mfilters*sizeof(UDM_FILTER));
		}
	}
	err=regcomp(&(Filter[nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err, &(Filter[nfilters].filter), regerrstr, ERRSTRSIZE);
		udm_log(0,UDM_LOG_ERROR,"Wrong regex in config file: %s: %s", filter,regerrstr);
		regfree(&(Filter[nfilters].filter));
		return(1);
	}
	Filter[nfilters].filter_type=filter_type;
	Filter[nfilters].reverse=reverse;
	Filter[nfilters].regstr=strdup(filter);
	nfilters++;
	return(0);
}
static void FreeFilters(){
int i;
	for(i=0; i < nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	free_it(Filter);
	nfilters=0;
	mfilters=0;
}

static char FCBuf[64]="";
char *FilterComStr(UDM_FILTER* f){
	FCBuf[0]=0;
	if(!f)return(FCBuf);
	switch(f->filter_type){
		case F_ALLOW: sprintf(FCBuf,"Allow");break;
		case F_DISALLOW: sprintf(FCBuf,"Disallow");break;
		case F_HEAD: sprintf(FCBuf,"CheckOnly");break;
		default: sprintf(FCBuf,"Unknown");break;
	}
	sprintf(STREND(FCBuf),"%s %s",
		(f->reverse?"NoMatch":""),
		(f->regstr?f->regstr:""));
	return(FCBuf);
}
UDM_FILTER * FindFilter(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<nfilters;i++){
		err=regexec(&(Filter[i].filter),url,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			return(&Filter[i]);
		if((!Filter[i].reverse)&&(!err))
			return(&Filter[i]);
	}
	return(NULL);
}



/******************** ALIASES ***************************************/

static int AddAlias(char * find,char *replace){
	if(naliases>=maliases){
		if(maliases){
			maliases+=16;
			Alias=(ALIAS *)realloc(Alias,maliases*sizeof(ALIAS));
		}else{
			maliases=16;
			Alias=(ALIAS *)malloc(maliases*sizeof(ALIAS));
		}
	}
	Alias[naliases].find=strdup(find);
	Alias[naliases].replace=strdup(replace);
	naliases++;
	return(0);
}
static void FreeAliases(){
int i;
	for(i=0;i<naliases;i++){
		free(Alias[i].find);
		free(Alias[i].replace);
	}
	naliases=0;
	maliases=0;
	free_it(Alias);
}

ALIAS * FindAlias(char * str){
int j;
	for(j=0;j<naliases;j++){
		if(!STRNCMP(str,Alias[j].find)){
			return(&Alias[j]);
		}
	}
	return(NULL);
}


/*************************** Servers *************************************/

static int AddServer(SERVER * srv){

	if(nservers>=mservers){
		if(mservers){
			mservers+=16;
			Server=(SERVER *)realloc(Server,mservers*sizeof(SERVER));
		}else{
			mservers=16;
			Server=(SERVER *)malloc(mservers*sizeof(SERVER));
		}
	}
	Server[nservers].url=strdup(srv->url);
	Server[nservers].charset=srv->charset?strdup(srv->charset):NULL;
	Server[nservers].basic_auth=srv->basic_auth?strdup(srv->basic_auth):NULL;
	Server[nservers].htdb_list=srv->htdb_list?strdup(srv->htdb_list):NULL;
	Server[nservers].htdb_doc=srv->htdb_doc?strdup(srv->htdb_doc):NULL;
	if(srv->proxy){
		char *s;
		Server[nservers].proxy=strdup(srv->proxy);
		if((s=strchr(Server[nservers].proxy,':'))){
			*s=0;
			Server[nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Server[nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Server[nservers].proxy=NULL;
		Server[nservers].proxy_port=0;
	}
	Server[nservers].period=srv->period;
	Server[nservers].hint=srv->hint;
	Server[nservers].net_errors=0;
	Server[nservers].outside=srv->outside;
	Server[nservers].maxhops=srv->maxhops;
	Server[nservers].gindex=srv->gindex;
	Server[nservers].gfollow=srv->gfollow;
	Server[nservers].deletebad=srv->deletebad;
	Server[nservers].userobots=srv->userobots;
	Server[nservers].bodyweight=srv->bodyweight;
	Server[nservers].titleweight=srv->titleweight;
	Server[nservers].urlweight=srv->urlweight;
	Server[nservers].descweight=srv->descweight;
	Server[nservers].keywordweight=srv->keywordweight;
	Server[nservers].max_net_errors=srv->max_net_errors;
	Server[nservers].read_timeout=srv->read_timeout;
	Server[nservers].delete_no_server=srv->delete_no_server;
	nservers++;
	return(0);
}
static void FreeServers(){
int i;
	for(i=0;i<nservers;i++){
		free_it(Server[i].url);
		free_it(Server[i].charset);
		free_it(Server[i].basic_auth);
		free_it(Server[i].htdb_list);
		free_it(Server[i].htdb_doc);
		free_it(Server[i].proxy);
	}
	nservers=mservers=0;
	free_it(Server);
}
SERVER * FindServer(char *url){
int i;
char *r;

	if(!Server)return(NULL);
	url=remove2dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<nservers;i++){
			if(!STRNCMP(url,Server[i].url)){
				*r='r';
				return(&Server[i]);
			}
		}
		*r='r';
	}else{
		for(i=0;i<nservers;i++)
			if(!STRNCMP(url,Server[i].url))
				return(&Server[i]);
	}
	return(0);
}
static int cmpserver(const void *s1,const void *s2){
	return(strlen(((SERVER*)s2)->url)-strlen(((SERVER*)s1)->url));
}
static int InitServer(SERVER * srv){
	srv->url=NULL;
	srv->proxy=NULL;
	srv->basic_auth=NULL;
	srv->charset=NULL;
	srv->htdb_list=NULL;
	srv->htdb_doc=NULL;
	srv->proxy_port=0;
	srv->period=DEFAULT_REINDEX_TIME;
	srv->hint=0;
	srv->net_errors=0;
	srv->max_net_errors=MAXNETERRORS;
	srv->read_timeout=READ_TIMEOUT;
	srv->outside=0;
	srv->delete_no_server=1;
	srv->maxhops=DEFAULT_MAX_HOPS;
	srv->gindex=1;
	srv->gfollow=1;
	srv->deletebad=0;
	srv->userobots=1;
	srv->bodyweight=1;
	srv->titleweight=2;
	srv->keywordweight=2;
	srv->descweight=2;
	srv->urlweight=0;	
	return(0);
}

/****************************  Load Configuration **********************/


__INDLIB__ int LoadConfig(char *conf_name,int config_level,int load_flags){
int i=0; /* For line number */
FILE *config;
char str[STRSIZ]="";
char str1[STRSIZ]="";
char str2[STRSIZ]="";
int has_dict=0;


	if(config_level==0){ /* Do some initialization */
		csrv=(SERVER*)malloc(sizeof(SERVER));
		InitServer(csrv);
		SetMaxURLNumber(-1);
		CurURLNumber=0;
		sprintf(UserAgent,"%s/%s",INDEXER_USER_AGENT,VERSION);
		ExtraHeaders[0]=0;
		MaxDocSize=MAXDOCSIZE;
		/*GMToffs=get_gmtoff();*/
#ifdef HAVE_SYSLOG_H
		LogFacility=LOG_FACILITY;
#endif
		Force1251=0;
		DBPort=0;
		free_it(DBHost);
		free_it(DBName);
		free_it(DBUser);
		free_it(DBPass);
		free_it(LocalCharset);
		LCindex=0;
		makeseparator(SEPARATOR,LCindex);
		FreeFilters();
		FreeAliases();
		free_it(MirrorRoot);
		free_it(MirrorHeadersRoot);
		use_mirror = UDM_MIRROR_NO;

		/* I have removed it to allow adding of URLs  */
		/* Passed in -i with -u command line args     */
		/* But it seems that UnloadConf() is required */
		/* REMOVED      FreeHrefs(); */

#ifdef USE_FILE
		FreeTypes();
#endif
		FreeServers();
		have_targets=1;
		index_started=0;
	}

	/* Open config file */
	if(!(config=fopen(conf_name,"r"))){
		udm_log(0,UDM_LOG_ERROR,"Error: can't open config file '%s': %s",conf_name, strerror(errno));
		return(1);
	}

	/*  Read lines and parse */
	while(fgets(str1,sizeof(str1),config)){
		char *end;
		i++;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		if(!STRNCASECMP(str,"DBHost")){
			if(sscanf(str+6,"%s",str1))
				DBHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBName")){
			DBName=strdup(trim_chars(str+6," \r\n\t"));
		}else
		if(!STRNCASECMP(str,"DBUser")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBPass")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBPass=strdup(str1);
			else	DBPass=strdup("");
		}else
		if(!STRNCASECMP(str,"DBPort")){
			if(sscanf(str+6,"%d",&DBPort)!=1)DBPort=0;
		}else
		if(!STRNCASECMP(str,"LocalCharset")){
			if(sscanf(str+13,"%s",str1)&&strcmp(str1,"")){
				free_it(LocalCharset);
				LocalCharset=strdup(str1);
				LCindex=getcharset(LocalCharset);
				makeseparator(SEPARATOR,LCindex);
			}
		}else
		if(!STRNCASECMP(str,"DBType")){
			if(sscanf(str+6,"%s",str1)){
				if(!STRNCASECMP(str1,"msql"))DBType=DB_MSQL;
				else
				if(!STRNCASECMP(str1,"solid"))DBType=DB_SOLID;
				else
				if(!STRNCASECMP(str1,"oracle"))DBType=DB_ORACLE;
				else
					DBType = DB_UNK;
			}else{
				DBType = DB_UNK;
			}
		}else
		if(!STRNCASECMP(str,"DisallowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_DISALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Disallow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_DISALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"AllowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_ALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Allow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_ALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnlyNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_HEAD,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnly")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_HEAD,0))
					return(1);
			}
		}else
#ifdef USE_FILE
		if(!STRNCASECMP(str,"AddType")){
			char *s,*s1,*lt;
			s=gettoken(str," \t\r\n",&lt);
			if((s1=gettoken(NULL," \t\r\n",&lt)))
				while((s=gettoken(NULL," \t\r\n",&lt)))
					AddType(s1,s);
		}else
#endif
		if(!STRNCASECMP(str,"Tag")){
			if(sscanf(str+4,"%d",&csrv->hint)!=1)
				csrv->hint=0;
		}else
		if(!STRNCASECMP(str,"MaxNetErrors")){
			if(sscanf(str+12,"%d",&csrv->max_net_errors)!=1)
				csrv->max_net_errors=MAXNETERRORS;
		}else
		if(!STRNCASECMP(str,"ReadTimeOut")){
			if(sscanf(str+11,"%d",&csrv->read_timeout)!=1)
				csrv->read_timeout=READ_TIMEOUT;
		}else
		if(!STRNCASECMP(str,"Period")){
			if(sscanf(str+6,"%d",&csrv->period)!=1)
				csrv->period=DEFAULT_REINDEX_TIME;
		}else
		if(!STRNCASECMP(str,"CharSet")){
			free_it(csrv->charset);
			if(sscanf(str+7,"%s",str1))
				csrv->charset=strdup(str1);
		}else
		if(!STRNCASECMP(str,"Proxy")){
			free_it(csrv->proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				csrv->proxy=strdup(str1);
		}else
		if(!STRNCASECMP(str,"HTTPHeader")){
			if(sscanf(str+11,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!STRNCMP(str1,"User-Agent: ")){
					strcpy(UserAgent,str1+12);
				}else{
					strcat(ExtraHeaders,str1);
					strcat(ExtraHeaders,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H
		if(!STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				LogFacility=syslog_facility(str1);
		}else
#endif
		if(!STRNCASECMP(str,"AuthBasic")){
			free_it(csrv->basic_auth);
			if(sscanf(str+9,"%s",str1)){
				csrv->basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,csrv->basic_auth,strlen(str1));
			}
		}else
		if(!STRNCASECMP(str,"HTDBList")){
			free_it(csrv->htdb_list);
			csrv->htdb_list=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"HTDBDoc")){
			free_it(csrv->htdb_doc);
			csrv->htdb_doc=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"Mime")){
			AddMime(str+4);
		}else
		if(!STRNCASECMP(str,"Alias")){
			if(sscanf(str+5,"%s%s",str1,str2)==2)
				AddAlias(str1,str2);
			else{
				udm_log(0,UDM_LOG_ERROR,
					"Error in config file '%s' line %d:%s",
					conf_name,i,str);
				return(1);
			}
		}else
		if(!STRNCASECMP(str,"Server")){
			URL from;
			if(sscanf(str+6,"%s",str1)!=1)continue;

			if(parseurl(&from,str1)){
				udm_log(0,UDM_LOG_ERROR,
					"File '%s' line %d: Invalid URL: %s",
					conf_name,i,str1);
				return(1);
			}
			csrv->url=strdup(str1);
			AddServer(csrv);
			if((!strcmp(from.schema,"http"))&&(load_flags&FLAG_ADD_SERV)&&(csrv->userobots)){
				sprintf(str,"%s://%s/%s",from.schema,from.hostinfo,"robots.txt");
				AddHref(NULL,str,0,0,0);
			}
			if(load_flags&FLAG_ADD_SERV)
				AddHref(NULL,str1,0,0,0);
		}else
		if(!STRNCASECMP(str,"FollowOutside")){
			if(sscanf(str+13,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->outside=1;
				else
					csrv->outside=0;
			}else
				csrv->outside=0;
		}else
		if(!STRNCASECMP(str,"Index")){
			if(sscanf(str+5,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->gindex=1;
				else
					csrv->gindex=0;
			}else
				csrv->gindex=1;
		}else
		if(!STRNCASECMP(str,"Follow ")){
			if(sscanf(str+7,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->gfollow=1;
				else
					csrv->gfollow=0;
			}else
				csrv->gfollow=1;
		}else
		if(!STRNCASECMP(str,"Robots")){
			if(sscanf(str+6,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->userobots=1;
				else
					csrv->userobots=0;
			}else
				csrv->userobots=1;
		}else
		if(!STRNCASECMP(str,"DeleteBad")){
			if(sscanf(str+9,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->deletebad=1;
				else
					csrv->deletebad=0;
			}else
				csrv->deletebad=0;
		}else
		if(!STRNCASECMP(str,"DeleteNoServer")){
			if(sscanf(str+14,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->delete_no_server=1;
				else
					csrv->delete_no_server=0;
			}else
				csrv->delete_no_server=1;
		}else
		if(!STRNCASECMP(str,"MaxHops")){
			if(sscanf(str+7,"%d",&csrv->maxhops)!=1)
				csrv->maxhops=0;
		}else
		if(!STRNCASECMP(str,"BodyWeight")){
			if(sscanf(str+10,"%d",&csrv->bodyweight)!=1)
				csrv->bodyweight=1;
		}else
		if(!STRNCASECMP(str,"TitleWeight")){
			if(sscanf(str+11,"%d",&csrv->titleweight)!=1)
				csrv->titleweight=2;
		}else
		if(!STRNCASECMP(str,"DescWeight")){
			if(sscanf(str+10,"%d",&csrv->descweight)!=1)
				csrv->descweight=2;
		}else
		if(!STRNCASECMP(str,"UrlWeight")){
			if(sscanf(str+9,"%d",&csrv->urlweight)!=1)
				csrv->urlweight=0;
		}else
		if(!STRNCASECMP(str,"KeywordWeight")){
			if(sscanf(str+13,"%d",&csrv->keywordweight)!=1)
				csrv->keywordweight=2;
		}else
		if(!STRNCASECMP(str,"MaxDocSize")){
			if(sscanf(str+11,"%d",&MaxDocSize)!=1)
				MaxDocSize=MAXDOCSIZE;
		}else
		if(!STRNCASECMP(str,"NoIndex")){
			csrv->gindex=0;
		}else
		if(!STRNCASECMP(str,"NoFollow")){
			csrv->gfollow=0;
		}else
		if(!STRNCASECMP(str,"ForceIISCharset1251")){
			if(sscanf(str+20,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))Force1251=1;
				else	Force1251=0;
			}else	Force1251=0;
		}else
		if(!STRNCASECMP(str,"Affix")){
			char lang[20];
			if(load_flags&&FLAG_SPELL)
			if(sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				if(ImportAffixes(lang,str)){
					udm_log(0,UDM_LOG_ERROR,
					"Can't load affix :%s",str);
					return(1);
				}
			}
		}else
		if(!STRNCASECMP(str,"Spell")){
			char lang[20];
			if(load_flags&&FLAG_SPELL)
			if(sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				/* -----  0 or 1 ?  */
				if(ImportDictionary(lang,str,0,"")){
					udm_log(0,UDM_LOG_ERROR,
					"Can't load dictionary :%s",str);
					return(1);
				}
				has_dict=1;
			}
		}else
		if(!STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				LoadConfig(str,config_level+1,load_flags);
			}
		}else
		if(!STRNCASECMP(str,"MirrorRoot")){
			if((sscanf(str+10,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/mirrors",CONFIGDIR);
			}
			MirrorRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		if(!STRNCASECMP(str,"MirrorHeadersRoot")){
			if((sscanf(str+17,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/headers",CONFIGDIR);
			}
			MirrorHeadersRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		if(!STRNCASECMP(str, "MirrorPeriod")){
			/* May be better make it useful for every Server? */
			if(sscanf(str + 12, "%d", &use_mirror) != 1)
				use_mirror = UDM_MIRROR_NO;
		}else{
			udm_log(0,UDM_LOG_ERROR,"Error in config file line %d: %s",i,str);
			return(1);
		}
		str[0]=0;
	}
	fclose(config);


	/* On level0 : Free some variables, prepare others, etc */
	if(config_level==0){
		/* Add one virtual server if we want FollowOutside */
		/* or DeleteNoServer no	*/
		
		if((csrv->outside)||(!csrv->delete_no_server)){
			free_it(csrv->url);
			csrv->url="";
			AddServer(csrv);
		}else{
			free_it(csrv->url);
		}
		free_it(csrv->charset);
		free_it(csrv->proxy);
		free_it(csrv->basic_auth);
		free_it(csrv->htdb_list);
		free_it(csrv->htdb_doc);
		/*  Long name should be found first    */
		/*  to allow different options         */
		/*  for server and it's subdirectories */
		qsort((void*)Server,nservers,sizeof(SERVER),cmpserver);
		/* Sort ispell dictionay if it has been loaded */
		if(has_dict)SortDictionary();
	}
	return(0);
}
